const Q            = require('q');
const fs           = require('fs');
const lib          = require('./functions');
const path         = require('path');
const spawn        = require('child_process').spawnSync
const request      = require('request');
const js2xmlparser = require("js2xmlparser");

const ERR_REQ  = 'REQUIRED_FIELDS';
const ERR_JSON = 'JSON_VALIDATION';
const ERR_PKG  = 'INTERNAL_PACKAGE_ERROR';

// todo

let RapidError = function(code, fields) {
    let messages = {
        'REQUIRED_FIELDS':        'Please, check and fill in required fields.',
        'JSON_VALIDATION':        'Syntax error. Incorrect input JSON. Please, check fields with JSON input.',
        'INTERNAL_PACKAGE_ERROR': 'Something went wrong inside the package.'
    }

    this.status_code = code;
    this.status_msg  = messages[code];

    if(code == 'REQUIRED_FIELDS')
        this.fields  = fields || [];
}

RapidError.prototype = Object.create(Error.prototype);
RapidError.prototype.constructor = RapidError;

class Api {

    /**
     * Init API.
     *
     * @param {String} options
     * @param {String} options.endpoint - Request API endpoint
     * @param {String} options.storage  - File tmp storage file
     * @param {Object} options.headers  - Custom http headers {'header': 'value'}
     */

    constructor(endpoint, options) {

        if(!endpoint || typeof endpoint !== 'string')
            throw new Error('Fatal: Api endpoint url is required.');

        options = options || {};

        this.headers  = options.headers || {};
        this.storage  = options.storage || ''; 
        this.endpoint = endpoint; 

        if (!this.headers["user-agent"])
            this.headers["user-agent"] = "RapidApi Client";

        this.types = {
            'NUMBER': (v) => {
                let parsed = parseInt(v);
                return isNaN(parsed) ? v : parsed; 
            },
            'JSON'  : (v) => this._json(v),
            'ARRAY' : (v) => this._json(v),
            'RAW'   : (v) => this._json(v)

            // 'FILE': (f) => { ... }
        }
    }

    /**
     * API auth
     *
     * @param {Object} options
     * @param {String} options.type     - Auth type: basic|bearer|...todo
     * @param {String} options.username - Username for basic auth
     * @param {String} options.password - Password for basic auth
     * @param {String} options.token    - Access Token
     */

    auth(options) {
        if(!options.type) return this;

        if(options.type == 'basic') {
            //if(!options.username) console.log('Warn: auth username is empty');
            //if(!options.password) console.log('Warn: auth password is empty');

            this.headers["Authorization"] = 'Basic ' + new Buffer(options.username + ':' + options.password).toString('base64');
        }

        if(options.type == 'bearer') {
            options = {
                'bearer': options.token
            }
        }

        this.auth = options;

        return this;
    }

    /**
     * API request
     *
     * @param {Object} options
     * @param {String} options.method    - request method. GET by default. POST if body is setted
     * @param {String} options.isForm    - true if API awaits for form data
     * @param {String} options.isRawBody - true if API awaits for raw post body
     * @param {String} options.parseUri  - true if you need replace uri params like /:uri with options.query keys
     * @param {String} options.brackets  - true if you need replace uri params like /{uri} with options.query keys
     * @param {String} options.parseMultiple - true if you need provide JSON->multiple_index1,multiple_index2 body
     * @param {String} options.hasSkip   - true if you need skip some request params
     * @param {Object} options.files     - key:file_link hash
     * @param {Object} options.body      - request body
     *
     * @return {Promise} Promise of request
     */

    request(options) {
        const defered = Q.defer();

        let method    = options.method || (options.body ? 'POST' : 'GET'),
            form      = options.isForm,
            rawBody   = options.isRawBody,
            parseUri  = options.parseUri,
            multiple  = options.parseMultiple,
            files     = options.files  || {},
            query     = options.query  || {},
            body      = options.body   || {},
            reqopts   = {};

        query = this._prepare(query);
        body  = this._prepare(body);

        if(multiple)
            body = this._parseMultiple(body);

        if(parseUri) this._parseUri(Object.assign({}, query, body), options.brackets);

        if(options.hasSkip) {
            query = this._skip(query);
            body  = this._skip(body);
        }

        query = lib.clearArgs(query);
        body  = lib.clearArgs(body, options.eo);

        body = body.__raw__ ? body.__raw__ : body;

        reqopts = {
            method: method,
            url:    this.endpoint,
            qs:     query
        }

        if(typeof this.auth !== 'function' && this.auth.type !== 'basic')    
            reqopts.auth = this.auth;

        if(this.headers) 
            reqopts.headers = this.headers

        if(options.hasTree)
            body = this._parseTree(body);

        if(rawBody) {
            this.headers['content-type'] = 'application/x-www-form-urlencoded';
            reqopts.body = this._rawBody(body);
        }

        if(form)
            reqopts.formData = body;

        if(!reqopts.body && (method == 'POST' || method == 'PUT')) {
            if(!options.xml) {
                this.headers['Content-Type'] = 'application/json';
                this.headers['Accept']       = 'application/json';
                this.headers['x-li-format']  = 'json';

                reqopts.json = true;
                reqopts.body = body;
            } else {
                this.headers['Content-Type'] = 'application/xml';
                reqopts.body = js2xmlparser.parse(options.xml.wrapper, body)
            }
        }

        if(Object.keys(files).length) {
            reqopts.formData = body;

            for(let file in files) {
                if(!files[file]) continue;

                let { 
                    stream, 
                    fpath 
                } = this._download(files[file]);

                fs.unlink(fpath, () => {});

                reqopts.formData[file] = {
                    value: stream,
                    options: {
                        filename: 'rapidapi' + lib.randomString
                    }
                };
            }

            delete reqopts.body;
        }

        reqopts.headers = this.headers;

        if(options.debug) console.log(reqopts);

        request(reqopts, (err, response, reslut) => {
            if(!err && (/20.*/).test(response.statusCode)) {
                if(options.includeRes)
                    defered.resolve(lib.safeParse({response, result: lib.safeParse(reslut)}));
                else 
                    defered.resolve(lib.safeParse(reslut));
            } else { 
                defered.reject(lib.safeParse(err || reslut || response.statusCode));
            }
        });

        return defered.promise;
    }

    /**
     * Fileds cast
     *
     * @param {String} _prop
     * @param {String} value
     *
     * @return {Object} casted request pair
     */

    _serialize(_prop, value) {
        let [prop, type] = _prop.split('|');
        
        type = type.toUpperCase();
        if(this.types[type]) 
            value = this.types[type](value);

        return {prop, value, type};
    }

    /**
     * Prepare fields
     * Required fields format: {'$key': 'value'}
     * Throws error if required fields is not filled
     *
     * @param {Object} object - request data
     *
     * @return {Object} clear request data
     */

    _prepare(object) {
        let _object = {},
            fill    = [];

        for(let _prop in object) {
            let {
                prop,
                value,
                type
            } = this._serialize(_prop, object[_prop]);

            if(type == 'RAW') _object['__raw__'] = value;

            if(prop[0] == '$') {
                prop = prop.replace('$', '');
                if(!value) {
                    prop = prop.replace('!', '');
                    fill[fill.length] = (/[-_]/.test(prop)) ? lib.toCamelCase(prop) : prop;
                }
            }

            _object[prop] = value;
        }

        if(fill.length > 0) throw new RapidError(ERR_REQ, fill);

        return _object;
    }

    /**
     * JSON object to multiple body keys convertation
     *
     * @param {Object} object - request data
     *
     * @return {Object} ready request data
     */

    _parseMultiple(object) {
        let muli = {};

        for(let prop in object) {
            if(!object[prop]) continue;

            if(/\*/.test(prop)) {
                try {
                    if(typeof muli[prop] !== 'object') muli[prop] = JSON.parse(object[prop]);
                    delete object[prop];
                } catch(e) {
                    throw new RapidError(ERR_JSON);
                }
            }
        }

        for(let key in muli) {
            for(let index in muli[key]) {
                let objkey = key.replace('*', index);
                object[objkey] = muli[key][index];
            }
        }

        return object;
    }

    /**
     * dotted.keys -> tree (content.submitted-url.facebook)
     *
     * @param {Object} object - request data
     *
     * @return {Object} ready request data
     */

    _parseTree(object) {
        for(let prop in object) {
            if(/\./.test(prop)) {
                let parts = prop.split('.');
                if(!object[parts[0]]) object[parts[0]] = {};

                for (var i = 1; i < parts.length; i++) {
                    object[parts[0]][parts[i]] = parts[i + 1] ? {} : object[prop];
                }

                delete object[prop];
            }
        }

        return object;
    }

     /**
     * Build raw post string
     *
     * @param {Object} body - request body
     *
     * @return {String} query string
     */

    _rawBody(body) {
        let bodyString = '';

        let encode = (str) => encodeURIComponent(str).replace(/[!'()*]/g, (c) => '%' + c.charCodeAt(0).toString(16));

        let serializeObj = (val, prop) => {
            for(let key in val) {
                if(typeof val[key] == 'object')
                    serializeObj(val[key], `${prop}[${key}]`)
                else 
                    bodyString += `&${prop}[${key}]=${encode(val[key])}`;
            }
        }

        for(let prop in body) {
            if(body[prop]) {
                if(typeof body[prop] == 'object') serializeObj(body[prop], prop);   
                else bodyString += `&${prop}=${encode(body[prop])}`; 
            }
        }

        return bodyString.slice(1);
    }

    _skip(object) {
        for(let prop in object) {
            if(prop[0] == '!') delete object[prop];
        }

        return object;
    }

    _parseUri(query, brackets) {
        for(let key in query) {
            let repkey = (brackets ? '{' + key + '}' : ':' + key).replace('!', '');
            this.endpoint = this.endpoint.replace(repkey, query[key])
        }
    }

    /**
     * Download a file
     *
     * @param {String} file - File
     *
     * @return {Stream} 
     */

    _download(file) {
        let attach = spawn(process.execPath, [require.resolve('./download.js'), file, this.storage]);
    
        if(!attach.stderr.toString()) {
            let response = JSON.parse(attach.stdout.toString());
            var fn       = path.resolve(this.storage, response.message);

            if(response.err) {
                throw new Error('Bad file!', res, {to});
                return;
            }

            return {
                fpath:  fn,
                stream: fs.createReadStream(fn),
            }

        } else {
            console.log('Error with download.js!', attach.stderr.toString());
            throw new RapidError(ERR_PKG)
        }
    }

    _json(str) {
        try {
           str = typeof str == 'string' 
                ? JSON.parse(str)
                : str;
        } catch(e) {
            throw new RapidError(ERR_JSON)
        }

        return str;
    }
}

module.exports = Api;
module.exports.Error = RapidError;